'''Assetto Corsa Websocket Server Class'''

import asyncio
import logging
import os
import websockets

from ac_websocket_server.debug import monitor_tasks
from ac_websocket_server.constants import HOST, PORT
from ac_websocket_server.error import WebsocketsServerError
from ac_websocket_server.game import GameServer
from ac_websocket_server.grid import Grid
from ac_websocket_server.handlers import handler
from ac_websocket_server.observer import Observer
from ac_websocket_server.protocol import Protocol

EXTRA_DEBUG = False


class WebsocketsServer(Observer):
    '''Represents an Assetto Corsa WebSocket Server.

    Allows control of an Assetto Corsa server with a websockets interface.'''
    # pylint: disable=logging-fstring-interpolation

    def __init__(self,
                 server_directory: str = None,
                 host: str = HOST,
                 port: int = PORT
                 ) -> None:

        self.__logger = logging.getLogger('ac-ws.ws-server')

        if EXTRA_DEBUG:
            asyncio.get_event_loop().create_task(monitor_tasks())

        self.connected = set()

        self.host = host
        self.port = port

        if not server_directory:
            self.server_directory = os.getcwd()
        else:
            self.server_directory = server_directory

        try:
            self.game = GameServer(server_directory=self.server_directory)
            self.game.subscribe(self)

            self.grid = Grid(server_directory=self.server_directory)
            self.grid.subscribe(self)
        except WebsocketsServerError as error:
            self.__logger.error(f'Fatal error {error}')
            raise

        self.stop_server: asyncio.Future = None

        self.__notifier_queue: asyncio.Queue = asyncio.Queue()

    async def consumer(self, message):
        '''ACWS consumer function for all received messages from client'''
        # pylint: disable=logging-fstring-interpolation, line-too-long

        if message == b'\n':
            return

        self.__logger.debug(f'Received message: {message}')

        message_words = str(message.strip(), 'utf-8').split()
        message_funcs = {'grid': self.grid.consumer,
                         'server': self.game.consumer}

        if message_funcs.get(message_words[0]):
            await message_funcs[message_words[0]](message_words[1:])
            return

        await self.__notifier_queue.put(Protocol.error(
            msg=f'Received unrecognised message: {message}'))

    async def handler(self, websocket):
        '''ACWS handler function for websocket connection'''

        self.connected.add(websocket)

        await websocket.send(Protocol.success(
            msg=f'Welcome to the Assetto Corsa WebSocker server running at {self.host}:{self.port}'))

        await handler(websocket, self.consumer, self.producer)

    async def notify(self, notifier):
        '''Receive a notification of a new message from
        notifier instance.  Pull the data off the notifier's
        queue and store in the local queue.'''

        message = await notifier.get()
        await self.__notifier_queue.put(message)

    async def producer(self):
        '''Pull a message off the queue and send on websocket'''
        data = await self.__notifier_queue.get()
        self.__logger.debug(f'Sending message: {data}')
        return data

    async def start(self):
        '''Start the websocket server'''

        try:

            self.__logger.info('Starting websocket server')

            self.stop_server = asyncio.Future()

            async with websockets.serve(self.handler, self.host, self.port):
                await self.stop_server

            self.__logger.info('Stopping websocket server')

        except KeyboardInterrupt:
            self.__logger.info('Interupting the server')

    async def stop(self):
        '''Stop the websocket server'''

        self.stop_server.set_result(True)



'''Assetto Corsa Grid Class'''

import glob
import logging
import os
from typing import List

from ac_websocket_server.entries import EntryList
from ac_websocket_server.error import WebsocketsServerError
from ac_websocket_server.observer import Notifier
from ac_websocket_server.protocol import Protocol


class Grid(Notifier):
    '''Represents grid setup'''

    def __init__(self, server_directory: str):

        self.__logger = logging.getLogger('ac-ws.grid')

        self.server_directory = server_directory
        entry_list_file_name = f'{self.server_directory}/cfg/entry_list.ini'

        if os.path.exists(entry_list_file_name):
            self.entry_list = EntryList(entry_list_file_name)
        else:
            error_message = f'Missing entry_list.ini file in {self.server_directory}'
            self.__logger.error(error_message)
            raise WebsocketsServerError(error_message)

        super().__init__()

    async def consumer(self, message_words: List[str]):
        '''Consume args destined for the server'''

        message_funcs = {'finish': self.__update_finishing,
                         'reverse': self.__update_reversed,
                         'current': self.__current,
                         'original': self.__entries,
                         'write': self.__write}

        if message_funcs.get(message_words[0]):
            await message_funcs[message_words[0]]()

    async def __current(self):
        '''Show current grid order info as a JSON string'''
        await self.put(Protocol.success({'grid': self.entry_list.show_grid()}))

    async def __entries(self):
        '''Show original grid order info as a JSON string'''
        await self.put(Protocol.success({'entries': self.entry_list.show_entries()}))

    async def update(self, reverse: bool = False):
        '''Update grid based on latest results.JSON file and using sorting rule'''

        try:
            result_files = glob.glob(
                os.path.join(f'{self.server_directory}/results', "*RACE.json"))
            result_file = max(result_files, key=os.path.getctime)
            self.entry_list.parse_result_file(result_file)
        except WebsocketsServerError as error:
            await self.put(Protocol.error(
                msg=f'{result_file} parse FAILURE: {error}'))
            return

        await self.put(Protocol.success(
            msg=f'{result_file} parse SUCCESS'))

        if reverse:
            self.entry_list.set_reversed_order()
        else:
            self.entry_list.set_standard_order()

        self.__current()

    async def __update_finishing(self):
        '''Update grid in standard finishing order'''
        await self.update()

    async def __update_reversed(self):
        '''Update grid in reverse finishing order'''
        await self.update(reverse=True)

    async def __write(self):
        '''Write updated entry_list.ini file'''
        try:
            self.entry_list.write(
                f'{self.server_directory}/cfg/entry_list.ini')
            await self.put(Protocol.success(msg='entry_list.ini file update SUCCESS'))
        except WebsocketsServerError as error:
            await self.put(Protocol.error(
                msg=f'entry_list.ini file update FAILURE: {error}'))



'''Generic observer/notifier super-class.'''

import asyncio
import logging

VERBOSE_DEBUG = False


class Observer:
    '''Super class for an Observer class.

    The observer must use notifer.subscribe() and
    notifier.unsubscribe() messages to ensure that
    notify() messages will be received.'''

    def __init__(self):
        '''Initialise an Observer with a local queue.'''

        self.__notifier_queue: asyncio.Queue = asyncio.Queue()

        super().__init__()

    async def notify(self, notifier):
        '''Receive a notification of a new message from
        notifier instance.  Pull the data off the notifier's
        queue and store in the local queue.'''

        message = await notifier.get()
        await self.__notifier_queue.put(message)


class Notifier:
    '''Super class for a Notifier class.

    The notifier receives subscribe() and unsubscribe()
    messages to register interested observers.

    The notifier sends a notify() to observers when
    data exists and then receives a get() to pop the
    data from queue.'''

    def __init__(self):
        '''Initialise a Notifier with a local queue.'''

        self.__logger = logging.getLogger('ac-ws.notifier')

        self.__observers = []
        self.__observer_queue: asyncio.Queue = asyncio.Queue()

        super().__init__()

    async def get(self):
        '''Fetch an item from the queue.  Returns None if the queue is empty.'''
        # pylint: disable=logging-fstring-interpolation

        try:
            response = self.__observer_queue.get_nowait()
        except asyncio.QueueEmpty:
            response = None

        if VERBOSE_DEBUG:
            self.__logger.debug(f'get(self) -> {response})')

        return response

    async def put(self, item):
        '''Put an item on the local queu and notify all observers.'''
        # pylint: disable=logging-fstring-interpolation

        if VERBOSE_DEBUG:
            self.__logger.debug(f'put(self, {item})')

        await self.__observer_queue.put(item)
        for obs in self.__observers:
            await obs.notify(self)

    def subscribe(self, observer):
        '''Subscribe an observer object for state changes.
        Observer object must include an async notify(self, observable, *args, **kwargs) method.'''
        self.__observers.append(observer)

    def unsubscribe(self, observer):
        '''Unsubscribe an observer object.'''
        try:
            self.__observers.remove(observer)
        except ValueError:
            pass


class ObserverNotifier(Observer, Notifier):
    '''Super class that is both an Observer and Notifier.'''

    def __init__(self):
        # pylint: disable=useless-super-delegation
        super().__init__()
